from ridi_oauth2.introspector.constants import JWKKeyType
from ridi_oauth2.introspector.dtos import JWKRSADto, JWKECDto, BaseJWKDto


class JWKDtoFactory:
    @staticmethod
    def get_dto(json) -> BaseJWKDto:
        kty = json.get('kty')
        if kty == JWKKeyType.RSA:
            return JWKRSADto(json)
        elif kty == JWKKeyType.EC:
            return JWKECDto(json)

        raise NotImplementedError



import traceback
from functools import wraps
from typing import Callable, Tuple, Type

_DEFAULT_RETRY_COUNT = 5


class RetryFailException(Exception):
    pass


def retry(
        retry_count: int = _DEFAULT_RETRY_COUNT,
        retriable_exceptions: Tuple[Type[BaseException]] = (BaseException,),
        print_stacktrace: bool = False,
):
    def _decorator(func: Callable):
        def _wrapper(*args, **kwargs):
            stacktraces = []

            for _ in range(0, retry_count):
                try:
                    value = func(*args, **kwargs)
                except Exception as e:
                    if is_retriable_exception(e, retriable_exceptions):
                        stacktraces.append(traceback.format_exc())
                        continue

                    raise e
                else:
                    return value

            if print_stacktrace:
                print('[RetryFail][StackTrace] %s', stacktraces)

            raise RetryFailException

        return wraps(func)(_wrapper)

    return _decorator


def is_retriable_exception(exception: Type[BaseException], retriable_exceptions: Tuple[Type[BaseException]]) -> bool:
    is_class = isinstance(exception, type)

    if is_class:
        return issubclass(exception, retriable_exceptions)

    return isinstance(exception, retriable_exceptions)



class InvalidJwtSignatureException(Exception):
    pass


class ExpireTokenException(Exception):
    pass


class InvalidToken(Exception):
    pass


class PublicKeyException(Exception):
    pass


class InvalidPublicKey(PublicKeyException):
    pass


class FailToLoadPublicKeyException(PublicKeyException):
    pass


class NotExistedKey(PublicKeyException):
    pass


class PublicRequestException(PublicKeyException):
    pass



import typing


class BaseIntrospector:
    def __init__(self, access_token: str, token_type_hint: str = None):
        self._access_token = access_token
        self._token_type_hint = token_type_hint

    @property
    def access_token(self) -> str:
        return self._access_token

    @property
    def token_type_hint(self) -> str:
        return self._token_type_hint

    def introspect(self) -> typing.Dict:
        raise NotImplementedError



from django.contrib.auth.base_user import AbstractBaseUser
from django.db import models

from ridi_django_oauth2.managers import RidiUserManager


class RidiUser(AbstractBaseUser):
    u_idx = models.IntegerField(primary_key=True, editable=False, verbose_name='u_idx')

    USERNAME_FIELD = 'u_idx'

    objects = RidiUserManager()

    class Meta:
        db_table = 'ridi_user'
        verbose_name = '사용자 계정'
        verbose_name_plural = '사용자 계정 리스트'

    def __str__(self):
        return str(self.get_username())



import typing
from base64 import urlsafe_b64decode
from datetime import datetime, timedelta

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric.ec import EllipticCurvePublicNumbers, SECP256R1
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicNumbers
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat

from ridi_django_oauth2_lib.utils.bytes import bytes_to_int
from ridi_oauth2.introspector.constants import JWK_EXPIRES_MIN, JWKCrv


class AccessTokenInfo:
    def __init__(self, subject: str, u_idx: int, expire: int, client_id: str, scope: typing.List):
        self._subject = subject
        self._u_idx = u_idx
        self._expire_timestamp = expire
        self._expire_date = datetime.fromtimestamp(expire)
        self._client_id = client_id
        self._scope = scope

    @property
    def subject(self) -> str:
        return self._subject

    @property
    def u_idx(self) -> int:
        return self._u_idx

    @property
    def expire_timestamp(self) -> int:
        return self._expire_timestamp

    @property
    def expire_date(self) -> datetime:
        return self._expire_date

    @property
    def client_id(self) -> str:
        return self._client_id

    @property
    def scope(self) -> typing.List:
        return self._scope

    @staticmethod
    def from_dict(dictionary: typing.Dict):
        return AccessTokenInfo(
            subject=dictionary['sub'], u_idx=dictionary['u_idx'], expire=dictionary['exp'], client_id=dictionary['client_id'],
            scope=dictionary['scope'],
        )


class BaseJWKDto:
    def __init__(self, json):
        self._json = json
        self.expires = datetime.now() + timedelta(minutes=JWK_EXPIRES_MIN)

    @property
    def kid(self) -> str:
        return self._json.get('kid')

    @property
    def kty(self) -> str:
        return self._json.get('kty')

    @property
    def use(self) -> str:
        return self._json.get('use')

    @property
    def alg(self) -> str:
        return self._json.get('alg')

    @property
    def is_expired(self) -> bool:
        return self.expires < datetime.now()


class JWKRSADto(BaseJWKDto):
    def __init__(self, json):
        super().__init__(json)
        decoded_n = bytes_to_int(urlsafe_b64decode(self.n))
        decoded_e = bytes_to_int(urlsafe_b64decode(self.e))
        rsa_public_key = RSAPublicNumbers(decoded_e, decoded_n).public_key(default_backend())
        self.public_key = rsa_public_key.public_bytes(Encoding.PEM, PublicFormat.SubjectPublicKeyInfo).decode()

    @property
    def e(self) -> str:
        return self._json.get('e')

    @property
    def n(self) -> str:
        return self._json.get('n')


class JWKECDto(BaseJWKDto):
    def __init__(self, json):
        super().__init__(json)
        decoded_x = bytes_to_int(urlsafe_b64decode(self.x))
        decoded_y = bytes_to_int(urlsafe_b64decode(self.y))
        ec_public_key = EllipticCurvePublicNumbers(
            decoded_x,
            decoded_y,
            self._get_curve_instance()
        ).public_key(default_backend())
        self.public_key = ec_public_key.public_bytes(Encoding.PEM, PublicFormat.SubjectPublicKeyInfo).decode()

    def _get_curve_instance(self):
        if self.crv == JWKCrv.P256:
            return SECP256R1()

        raise NotImplementedError

    @property
    def crv(self) -> str:
        return self._json.get('crv')

    @property
    def x(self) -> str:
        return self._json.get('x')

    @property
    def y(self) -> str:
        return self._json.get('y')



from django.contrib.auth.base_user import BaseUserManager

from ridi_django_oauth2.exceptions import ImpossibleSetSuperuserException


class RidiUserManager(BaseUserManager):
    def create_user(self, u_idx: int, **kwargs):
        ridi_user = self.model(u_idx=u_idx, **kwargs)
        ridi_user.set_unusable_password()
        ridi_user.save()

        return ridi_user

    def create_superuser(self, u_idx: int, **kwargs):
        raise ImpossibleSetSuperuserException



from typing import Callable, Optional

from django.conf import settings


class _SettingKeyName:
    KEY_URL = 'RIDI_OAUTH2_KEY_URL'
    COOKIE_DOMAIN = 'RIDI_OAUTH2_COOKIE_DOMAIN'
    ACCESS_TOKEN_COOKIE_KEY = 'RIDI_OAUTH2_ACCESS_TOKEN_COOKIE_KEY'
    REFRESH_TOKEN_COOKIE_KEY = 'RIDI_OAUTH2_REFRESH_TOKEN_COOKIE_KEY'
    GET_USER_FROM_TOKEN_INFO = 'RIDI_OAUTH2_GET_USER_FROM_TOKEN_INFO'


class _Default:
    COOKIE_DOMAIN = 'ridibooks.com'
    ACCESS_TOKEN_COOKIE_KEY = "ridi-at"
    REFRESH_TOKEN_COOKIE_KEY = "ridi-rt"


# Cookie
_RIDI_COOKIE_DOMAIN = getattr(settings, _SettingKeyName.COOKIE_DOMAIN, _Default.COOKIE_DOMAIN)
_RIDI_ACCESS_TOKEN_COOKIE_KEY = getattr(settings, _SettingKeyName.ACCESS_TOKEN_COOKIE_KEY, _Default.ACCESS_TOKEN_COOKIE_KEY)
_RIDI_REFRESH_TOKEN_COOKIE_KEY = getattr(settings, _SettingKeyName.REFRESH_TOKEN_COOKIE_KEY, _Default.REFRESH_TOKEN_COOKIE_KEY)

_RIDI_OAUTH2_KEY_URL = getattr(settings, _SettingKeyName.KEY_URL)

_RIDI_OAUTH2_GET_USER_FROM_TOKEN_INFO = getattr(settings, _SettingKeyName.GET_USER_FROM_TOKEN_INFO, None)


class RidiOAuth2Config:
    @staticmethod
    def get_key_url() -> str:
        return _RIDI_OAUTH2_KEY_URL

    @staticmethod
    def get_cookie_domain() -> str:
        return _RIDI_COOKIE_DOMAIN

    @staticmethod
    def get_access_token_cookie_key() -> str:
        return _RIDI_ACCESS_TOKEN_COOKIE_KEY

    @staticmethod
    def get_refresh_token_cookie_key() -> str:
        return _RIDI_REFRESH_TOKEN_COOKIE_KEY

    @staticmethod
    def get_user_from_token_info_callable() -> Optional[Callable]:
        return _RIDI_OAUTH2_GET_USER_FROM_TOKEN_INFO if callable(_RIDI_OAUTH2_GET_USER_FROM_TOKEN_INFO) else None



from setuptools import find_packages, setup

version = '1.0.7'

# When the project is installed by pip, this is the specification that is used to install its dependencies.
install_requires = [
    "djangorestframework>=3.7.7",
    "django>=2.2.1",
    "PyJWT==1.6.1",
    "requests>=2.20.0",
    "cryptography>=2.3",
]

setup(
    name='ridi-django-oauth2',
    version=version,
    url='https://github.com/ridi/django-oauth2',
    license='MIT',
    description='RIDI Django OAuth2 Library',
    keywords=['oauth2', 'oauth2-resource', 'ridi', 'ridibooks'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',

        'License :: OSI Approved :: MIT License',
        'Framework :: Django :: 2.0',
        'Programming Language :: Python :: 3.6',
    ],
    packages=find_packages(exclude=["tests.*", "tests"]),
    install_requires=install_requires,
    include_package_data=True,
)



import typing
import urllib.parse
from json import JSONDecodeError

import requests

from ridi_oauth2.client.constants import OAuth2GrantType
from ridi_oauth2.client.dtos import AuthorizationServerInfo, ClientInfo, TokenData
from ridi_oauth2.client.exceptions import InvalidResponseException, OAuthFailureException
from ridi_oauth2.common.constants import HttpMethod


class Grant:
    def __init__(
            self, client_info: ClientInfo, auth_server_info: AuthorizationServerInfo
    ):
        self.client_info = client_info
        self.auth_server_info = auth_server_info

    def authorize(self, state: str) -> str:
        query = urllib.parse.urlencode({
            'client_id': self.client_info.client_id,
            'redirect_uri': self.client_info.redirect_uri,
            'scope': self.client_info.scope,
            'state': state,
            'response_type': 'code',
        })
        authorize_url = '{}?{}'.format(self.auth_server_info.authorization_url, query)
        return authorize_url

    def code(self, code: str) -> TokenData:
        data = self._get_default_token_data(OAuth2GrantType.AUTHORIZATION_CODE)
        data['code'] = code
        data['redirect_uri'] = self.client_info.redirect_uri

        return self._request_token(data=data)

    def refresh(self, refresh_token: str) -> TokenData:
        data = self._get_default_token_data(OAuth2GrantType.REFRESH_TOKEN)
        data['refresh_token'] = refresh_token

        if self.client_info.scope:
            data['scope'] = self.client_info.scope

        return self._request_token(data=data)

    def _get_default_token_data(self, grant_type: str) -> typing.Dict:
        return {
            'grant_type': grant_type,
            'client_id': self.client_info.client_id,
            'client_secret': self.client_info.client_secret,
        }

    def _request_token(self, data: typing.Dict) -> TokenData:
        return self._request(
            method=HttpMethod.POST, url=self.auth_server_info.token_url, data=data, headers={'Accept': 'application/json', }
        )

    @classmethod
    def _request(cls, method: str, url: str, data: typing.Dict, headers: typing.Dict = None) -> TokenData:
        try:
            response = requests.request(method=method, url=url, data=data, headers=headers)
            response.raise_for_status()
            return TokenData.from_dict(response.json())
        except JSONDecodeError as e:
            raise InvalidResponseException() from e
        except requests.HTTPError as e:
            cls._process_exception(exception=e)

    @staticmethod
    def _process_exception(exception: requests.HTTPError):
        response = exception.response

        try:
            error_response = response.json()
        except JSONDecodeError as e:
            raise exception from e

        error = error_response.get('error', None)
        error_description = error_response.get('error_description', None)
        error_uri = error_response.get('error_uri', None)

        error_exception = OAuthFailureException(error_code=error, description=error_description, error_uri=error_uri)
        raise error_exception



import typing

from django.http import HttpResponse, HttpResponseForbidden

from ridi_django_oauth2.response import HttpUnauthorizedResponse
from ridi_oauth2.resource.scope_checker import ScopeChecker

RESPONSE_HANDLER_TYPE = typing.Optional[typing.Callable]


def login_required(response_handler: RESPONSE_HANDLER_TYPE = None):
    def decorator(func):
        def wrapper(self, request, *args, **kwargs):
            user = request.user

            if not user or not user.is_authenticated:
                return _process_response_handler(request, response_handler, *args, **kwargs) or HttpUnauthorizedResponse()

            token_info = getattr(request.user, 'token_info', None)
            if token_info is None:
                return _process_response_handler(request, response_handler, *args, **kwargs) or HttpUnauthorizedResponse()

            return func(self, request, *args, **kwargs)
        return wrapper
    return decorator


def scope_required(required_scopes: typing.List, response_handler: RESPONSE_HANDLER_TYPE = None):
    def decorator(func):
        def wrapper(self, request, *args, **kwargs):
            token_info = request.user.token_info

            if not ScopeChecker.check(require_scopes=required_scopes, user_scopes=token_info.scope):
                return _process_response_handler(
                    request, response_handler, required_scopes=required_scopes, *args, **kwargs
                ) or HttpResponseForbidden()

            return func(self, request, *args, **kwargs)
        return login_required(response_handler)(wrapper)
    return decorator


def _process_response_handler(request, response_handler: RESPONSE_HANDLER_TYPE, *args, **kwargs) -> typing.Optional[HttpResponse]:
    if response_handler is None:
        return None

    return response_handler(request, *args, **kwargs)
